import random

collaborators = ''

# Enter a float in the interval [0.0, 1.0] for each part 1 of problem 1.
# Enter True or False for each part 2.

"""
Notice that none of these are universal hash families, since if m is what we are modding by,
then we have that for all i: h_i(x) = h_i(x + m)
Universal hash constructions are typically more complicated.
"""

answer_for_problem_1_part_a_1 = 1/2.0
answer_for_problem_1_part_a_2 = False 
answer_for_problem_1_part_b_1 = 4/9.0 
answer_for_problem_1_part_b_2 = False 
answer_for_problem_1_part_c_1 = 11/12.0
answer_for_problem_1_part_c_2 = False
answer_for_problem_1_part_d_1 = 6/7.0
answer_for_problem_1_part_d_2 = False

# On problem 2, enter an integer for parts a and b.
# For parts c and d, enter a string proving that no such sequence exists,
# or an insertion sequence providing a counter-example, not both.
answer_for_problem_2_part_a = 3
answer_for_problem_2_part_b = 4

"""
Here is a general solution:
desired is an array, with what you want in slot i of the open addressing table in desired[i]
"""

def get_sequence(desired):
  sequence = [('i', 100 + x) for x in xrange(10)]
  for i in range(10):
    sequence.append(('d', 100 + i))
    sequence.append(('i', desired[i]))
  return sequence

answer_for_problem_2_part_c = get_sequence([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
answer_for_problem_2_part_d = get_sequence([9, 8, 7, 6, 5, 4, 3, 2, 1, 0])

"""
Here are some shorter sequences that also work
"""
answer_for_problem_2_part_c = [('i', 0), ('i', 11), ('i',12), ('i', 13), ('i', 14), ('i', 15),\
                               ('i', 16), ('i', 17), ('i', 18), ('i', 10), ('d', 18), ('i', 9),\
                               ('d', 17), ('i', 8), ('d', 16), ('i', 7), ('d', 15), ('i', 6), \
                               ('d', 14), ('i', 5), ('d', 13), ('i', 4), ('d', 12), ('i', 3), \
                               ('d', 11), ('i', 2), ('d', 0), ('i', 1)]
answer_for_problem_2_part_d = [('i', 19), ('i', 9), ('i',18), ('i', 8), ('i', 17), ('i', 7),\
                               ('i', 16), ('i', 6), ('i', 15), ('i', 5), ('d', 19), ('d', 18),\
                               ('d', 17), ('d', 16),('d', 15), ('i', 4),('i', 3),('i', 2),('i', 1),('i',0)]


# Enter your answer to problem 3 here. 
answer_for_problem_3 = '''
Algorithm:
  Make a new array of size k, which contains the updated entries p_(i_j) + d_j for j in range(1,k+1).
  While doing so, mark the changed prices in the original array with a special 'Changed' marker.  This all takes O(k). 
  Sort the new array, in O(k log k)
  Merge the new array and the original array, using the same algorithm as that from merge-sort, 
     but skip the finger in the original array over entries with the 'Changed' marker.  This takes O(n).
  Return this merged array

The algorithm is clearly correct, and runs in O(k log k + n).

An informal, non-rigorous argument that this is optimal:
We can't do better than O(k log k), since we need to sort k new prices, which could be arbitrary. 
We can't do better than O(n), since if the smallest price was updated to the middle, and nothing else changed,
we would still need to shift at least n/2 elements in the array 
Thus we can't do better than O(max(k log k, n)) = O(k log k + n)
'''

""" 
Our original solution to problem 4, which was actually designed for 
the problem of correcting one-off polyominoes, instead of bitstrings
"""

# First, preprocess a bunch of random 64 bit integers r_i, each of which corresponds to one position in a messgae
r = [random.randint(1, (2**63) - 1) for i in xrange(50000)]
roll = [r[0]] + [r[i-1] ^ r[i] for i in xrange(1, 50000)]

def recover_original_messages(valid_messages, corrupted_messages):
  (n, k) = (len(valid_messages[0]), len(valid_messages))

  # Our hash function is the xor of the r_i's where message[i] is 1 
  # This hash is easy to roll
  def myhash(message):
    return reduce(lambda x, y : x ^ y, (r[i] for i in xrange(n) if message[i] == '1'), 0)

  valid_hash = {}
  for i in xrange(k):
    valid_hash[myhash(valid_messages[i])] = i
    
  answer = []
  for cor in corrupted_messages:
    hash = myhash(cor)
    for j in xrange(n):
      hash ^= roll[j]
      if hash in valid_hash:
        answer.append(valid_hash[hash])
        break
      
  return answer

"""
Here's a time-optimized solution which gets ~0.20 seconds total on the large test cases
Code stolen (with permission) from Joshua Blum, Tal Tchwella, and Rishikesh Tirumala and modified.
I'm sure many others had similar code, and it's possible there were faster things out there 
"""

def recover_original_messages(valid_messages, corrupted_messages):
  (n, k, c) = (len(valid_messages[0]), len(valid_messages), len(corrupted_messages))

  answer = [0] * c 

  valid_hash = {}
  for i in xrange(k):
    valid_hash[int(valid_messages[i], 2)] = i 

  corrupt_hash = {}
  for i in xrange(c):
    corrupt_hash[int(corrupted_messages[i],2)] = i

  if (n < 5000): # This part due to Rishikesh Tirumala

    pows = [(1 << i) for i in xrange(n)]
    for c_hash in corrupt_hash:
      for i in xrange(n):
        hash = c_hash ^ pows[i] # Flip bit i 
        if hash in valid_hash:
          answer[corrupt_hash[c_hash]] = valid_hash[hash]
          break

  else: # This part due to Joshua Blum, Tal Tchwella

    for c_hash in corrupt_hash:
      for v_hash in valid_hash:
        xor = c_hash ^ v_hash # bitwise xor should have exactly one 1
        if not (xor & (xor-1)): # check if xor is a power of two
          answer[corrupt_hash[c_hash]] = valid_hash[v_hash]
          break

  return answer 
